from xchainpy_util.asset import Asset

class Coin:
    def __init__(self, asset: Asset, amount):
        self._asset = asset
        self._amount = amount

    @property
    def asset(self):
        return self._asset

    @asset.setter
    def asset(self, asset):
        self._asset = asset

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, amount):
        self._amount = amount


import pytest
from xchainpy_binance import utils


class TestBinanceUtils:

    def test_get_prefix_testnet(self):
        assert utils.get_prefix('testnet') == 'tbnb'

    def test_get_prefix_mainnet(self):
        assert utils.get_prefix('mainnet') == 'bnb'

    def test_get_tx_type_transfer(self):
        assert utils.get_tx_type('TRANSFER') == 'transfer'

    def test_get_tx_type_deposit(self):
        assert utils.get_tx_type('DEPOSIT') == 'transfer'

    def test_get_tx_type_others(self):
        assert utils.get_tx_type('something') == 'unknown'

    def test_parse_tx(self):
        origin_tx = {
            'txHash': '0C6B721844BB5751311EC8910ED17F6E950E7F2D3D404145DBBA4E8B6428C3F1',
            'blockHeight': 123553830,
            'txType': 'TRANSFER',
            'timeStamp': '2020-11-03T17:21:34.152Z',
            'fromAddr': 'bnb1jxfh2g85q3v0tdq56fnevx6xcxtcnhtsmcu64m',
            'toAddr': 'bnb1c259wjqv38uqedhhufpz7haajqju0t5thass5v',
            'value': '4.97300000',
            'txAsset': 'USDT-6D8',
            'txFee': '0.00037500',
            'proposalId': None,
            'txAge': 58638,
            'orderId': None,
            'code': 0,
            'data': None,
            'confirmBlocks': 0,
            'memo': '',
            'source': 0,
            'sequence': 1034585,
        }
        tx = utils.parse_tx(origin_tx)
        assert tx.tx_from[0].address == origin_tx['fromAddr']
        assert tx.tx_to[0].address == origin_tx['toAddr']



import pytest
from xchainpy_client.models.types import Network, XChainClientParams
from xchainpy_binance.client import Client
from xchainpy_util.asset import AssetBNB
from xchainpy_client.models import tx_types

class TestBinanceClient:

    # Note: This phrase is created by https://iancoleman.io/bip39/ and will never been used in a real-world
    phrase = 'rural bright ball negative already grass good grant nation screen model pizza'
    mainnetaddress_path0 = 'bnb1zd87q9dywg3nu7z38mxdcxpw8hssrfp9e738vr'
    mainnetaddress_path1 = 'bnb1vjlcrl5d9t8sexzajsr57taqmxf6jpmgng3gmn'

    testnetaddress_path0 = 'tbnb1zd87q9dywg3nu7z38mxdcxpw8hssrfp9htcrvj'
    testnetaddress_path1 = 'tbnb1vjlcrl5d9t8sexzajsr57taqmxf6jpmgaacvmz'

    phraseForTX = 'wheel leg dune emerge sudden badge rough shine convince poet doll kiwi sleep labor hello'
    testnetaddress_path0ForTx = 'tbnb1t95kjgmjc045l2a728z02textadd98yt339jk7'
    
    transfer_amount = 0.0001


    @pytest.fixture
    def client(self):
        self.client = Client(XChainClientParams(network=Network.Mainnet, phrase=self.phrase))
        yield
        self.client.purge_client()

    def test_empty_wallet_main(self):
        client = Client(XChainClientParams(network=Network.Mainnet, phrase=self.phrase))
        address_main = client.get_address()
        assert address_main == self.mainnetaddress_path0

    def test_empty_wallet_test(self):
        client = Client(XChainClientParams(network=Network.Testnet, phrase=self.phrase))
        address_test = client.get_address()
        assert address_test == self.testnetaddress_path0

    def test_invalid_phrase(self):
        with pytest.raises(Exception) as err:
            assert Client(XChainClientParams(network=Network.Testnet, phrase='invalid phrase'))
        assert str(err.value) == "invalid phrase"

    def test_update_net(self, client):
        self.client.set_network('testnet')
        assert self.client.get_network() == 'testnet'
        assert self.client.get_address() == self.testnetaddress_path0

    def test_right_address(self, client):
        assert self.client.get_address() == self.mainnetaddress_path0
        assert self.client.get_address(1) == self.mainnetaddress_path1
        self.client.set_network('testnet')
        assert self.client.get_address() == self.testnetaddress_path0
        assert self.client.get_address(1) == self.testnetaddress_path1

    def test_set_phrase_return_address(self, client):
        assert self.client.set_phrase(self.phrase) == self.mainnetaddress_path0
        self.client.set_network('testnet')
        assert self.client.set_phrase(self.phrase) == self.testnetaddress_path0

    def test_validate_address(self, client):
        assert self.client.validate_address(self.mainnetaddress_path0) == True

    def test_validate_address_false_address(self, client):
        assert self.client.validate_address(self.mainnetaddress_path0 + '1') == False

    @pytest.mark.asyncio
    async def test_has_no_balances(self, client):
        assert await self.client.get_balance(address='bnb1v8cprldc948y7mge4yjept48xfqpa46mmcrpku') == []

    @pytest.mark.asyncio
    async def test_has_balances(self, client):
        self.client.set_network('testnet')
        assert await self.client.get_balance(address=self.testnetaddress_path0)

    @pytest.mark.asyncio
    async def test_balance_has_correct_asset(self, client):
        self.client.set_network('testnet')
        balance = await self.client.get_balance(self.testnetaddress_path0, AssetBNB)
        assert str(balance[0].asset) == str(AssetBNB)
        assert balance[0].amount

    @pytest.mark.asyncio
    async def test_get_transfer_fees(self, client):
        self.client.set_network('testnet')
        fee = await self.client.get_fees()
        assert fee.average
        assert fee.fast
        assert fee.fastest

    @pytest.mark.asyncio
    async def test_get_multi_send_fees(self, client):
        fee = await self.client.get_multi_send_fees()
        assert fee.average
        assert fee.fast
        assert fee.fastest

    @pytest.mark.asyncio
    async def test_get_single_and_multi_fees(self, client):
        fee = await self.client.get_single_and_multi_fees()
        assert fee['multi'].average 
        assert fee['multi'].fast
        assert fee['multi'].fastest
        assert fee['single'].average
        assert fee['single'].fast
        assert fee['single'].fastest

    @pytest.mark.asyncio
    async def test_should_broadcast_transfer(self, client):
        self.client.set_network('testnet')
        self.client.set_phrase(self.phraseForTX)
        assert self.client.get_address() == self.testnetaddress_path0ForTx
        before_balance = await self.client.get_balance(self.testnetaddress_path0ForTx)
        assert len(before_balance) == 1
        before_balance_amount = before_balance[0].amount
        await self.client.transfer(tx_types.TxParams(asset=AssetBNB, amount=self.transfer_amount, recipient=self.testnetaddress_path0ForTx))
        after_balance = await self.client.get_balance(self.testnetaddress_path0ForTx)
        after_balance_amount = after_balance[0].amount
        # assert round((float(before_balance_amount) - float(after_balance_amount)) * 10**8) == round((await self.client.get_fees()).average * 10 ** 8)

    @pytest.mark.asyncio
    async def test_should_raise_exception_if_input_amount_higher_than_balance(self , client):
        self.client.set_network('testnet')
        self.client.set_phrase(self.phraseForTX)
        before_balance = await self.client.get_balance(self.testnetaddress_path0ForTx)
        before_balance_amount = before_balance[0].amount
        send_amount = float(before_balance_amount) + 1
        with pytest.raises(Exception) as err:
            await self.client.transfer(tx_types.TxParams(asset=AssetBNB, amount=send_amount, recipient=self.testnetaddress_path0ForTx))
        assert str(err.value) == "input asset amout is higher than current (asset balance - transfer fee)"

    @pytest.mark.asyncio
    async def test_has_empty_tx_history(self, client):
        self.client.set_network('testnet')
        address = self.client.set_phrase('nose link choose blossom social craft they better render provide escape talk')
        params = tx_types.TxHistoryParams(address=address, limit=1)
        transactions = await self.client.get_transactions(params)
        assert transactions
        assert len(transactions.txs) == 0

    @pytest.mark.asyncio
    async def test_get_transactions(self, client):
        self.client.set_network('testnet')
        params = tx_types.TxHistoryParams(address=self.testnetaddress_path0ForTx, limit=1)
        transactions = await self.client.get_transactions(params)
        assert transactions
        assert len(transactions.txs) == 1 or 0
        if transactions.total > 0:
            assert isinstance(transactions.txs[0], tx_types.TX)

    @pytest.mark.asyncio
    async def test_get_transaction_data(self, client):
        self.client.set_network('testnet')
        try:
            params = tx_types.TxHistoryParams(address=self.testnetaddress_path0ForTx, limit=1)
            transactions = await self.client.get_transactions(params)
            transaction = await self.client.get_transaction_data(transactions.txs[0].tx_hash)
            assert transaction
            assert isinstance(transaction, tx_types.TX)
        except Exception as err:
            assert str(err) == 'list index out of range' # there is not any transaction in the last 3 months for this address

    def test_should_return_valid_explorer_url(self, client):
        assert self.client.get_explorer_url() == 'https://explorer.binance.org'
        self.client.set_network(Network.Testnet)
        assert self.client.get_explorer_url() == 'https://testnet-explorer.binance.org'

    def test_should_retrun_valid_explorer_address_url(self, client):
        assert self.client.get_explorer_address_url('anotherTestAddressHere') == 'https://explorer.binance.org/address/anotherTestAddressHere'
        self.client.set_network(Network.Testnet)
        assert self.client.get_explorer_address_url('testAddressHere') == 'https://testnet-explorer.binance.org/address/testAddressHere'

    def test_should_retrun_valid_explorer_tx_url(self, client):
        assert self.client.get_explorer_tx_url('anotherTestTxHere') == 'https://explorer.binance.org/tx/anotherTestTxHere'
        self.client.set_network(Network.Testnet)
        assert self.client.get_explorer_tx_url('testTxHere') == 'https://testnet-explorer.binance.org/tx/testTxHere'

    # @pytest.mark.asyncio
    # async def test_multi_send(self, client):
    #     self.client.set_network('testnet')
    #     self.client.set_phrase(self.phraseForTX)
    #     assert self.client.get_address() == self.testnetaddressForTx
    #     before_balance = await self.client.get_balance()
    #     assert len(before_balance) == 1
    #     before_balance_amount = before_balance[0].amount

    #     coins = [Coin(AssetBNB, self.transfer_amount), Coin(Asset.from_str('BNB.BNB'), self.transfer_amount)]
    #     tx_hash = await self.client.multi_send(coins, 'tbnb185tqzq3j6y7yep85lncaz9qeectjxqe5054cgn')

    #     after_balance = await self.client.get_balance()
    #     after_balance_amount = after_balance[0].amount
    #     assert round((float(before_balance_amount) - float(after_balance_amount)) * 10**8) == round(self.multi_tx_fee * 10 ** -8, 8)


class Fee:
    def __init__(self, msg_type:str, fee, fee_for:int):
        self._msg_type = msg_type
        self._fee = fee
        self._fee_for = fee_for

    @property
    def msg_type(self):
        return self._msg_type

    @msg_type.setter
    def msg_type(self, msg_type):
        self._msg_type = msg_type

    @property
    def fee(self):
        return self._fee

    @fee.setter
    def fee(self, fee):
        self._fee = fee

    @property
    def fee_for(self):
        return self._fee_for

    @fee_for.setter
    def fee_for(self, fee_for):
        self._fee_for = fee_for


class TransferFee:
    def __init__(self, fixed_fee_params:Fee, multi_transfer_fee:int, lower_limit_as_multi:int):
        self._fixed_fee_params = fixed_fee_params
        self._multi_transfer_fee = multi_transfer_fee
        self._lower_limit_as_multi = lower_limit_as_multi

    @classmethod
    def from_binance_fee(cls, transfer_fee):
        """Get TransferFee object from a binance fee

        :param transfer_fee: binance fee
        :type transfer_fee: dict
        :returns: TransferFee object
        """
        return TransferFee(fixed_fee_params=Fee(msg_type=transfer_fee['fixed_fee_params']['msg_type'],
                                                fee=transfer_fee['fixed_fee_params']['fee'],
                                                fee_for=transfer_fee['fixed_fee_params']['fee_for']),
                            multi_transfer_fee=transfer_fee['multi_transfer_fee'],
                            lower_limit_as_multi=transfer_fee['lower_limit_as_multi'])
    
    @property
    def fixed_fee_params(self):
        return self._fixed_fee_params

    @fixed_fee_params.setter
    def fixed_fee_params(self, fixed_fee_params):
        self._fixed_fee_params = fixed_fee_params

    @property
    def multi_transfer_fee(self):
        return self._multi_transfer_fee

    @multi_transfer_fee.setter
    def multi_transfer_fee(self, multi_transfer_fee):
        self._multi_transfer_fee = multi_transfer_fee

    @property
    def lower_limit_as_multi(self):
        return self._lower_limit_as_multi

    @lower_limit_as_multi.setter
    def lower_limit_as_multi(self, lower_limit_as_multi):
        self._lower_limit_as_multi = lower_limit_as_multi


from xchainpy_util.asset import Asset
from xchainpy_client.models.balance import Balance
from xchainpy_util.chain import Chain

class BinanceBalance(Balance):

    def __init__(self, balance):
        """
        :param balance: binance balance object
        :type balance: a dict contains: free, frozen, locked, symbol
        """
        asset = Asset(Chain.Binance.value, balance['symbol'])
        super().__init__(asset, balance['free'])


from xchainpy_util.asset import Asset
from xchainpy_client.models.balance import Balance
from xchainpy_util.chain import BNBCHAIN
from xchainpy_client.models import tx_types
import datetime


def get_prefix(network):
    """Convert network type to prefix of address

    :param network: testnet or mainnet
    :type network: str
    :returns: tbnb or bnb
    """
    return 'tbnb' if network == 'testnet' else 'bnb'


def get_tx_type(tx_type: str):
    """Get tx type

    :param tx_type: type of transaction
    :type tx_type: str
    :returns: transfer or unknown
    """

    if tx_type == 'TRANSFER' or tx_type == 'DEPOSIT':
        return 'transfer'
    return 'unknown'


def parse_tx(tx):
    """Parse tx

    :param tx: The transaction to be parsed
    :type tx: str
    :returns: The transaction parsed from the binance tx
    """
    asset = Asset.from_str(f'BNB.{tx["txAsset"]}')
    tx_from = [tx_types.TxFrom(tx['fromAddr'], tx['value'])]
    tx_to = [tx_types.TxTo(tx['toAddr'], tx['value'])]
    tx_date = datetime.datetime.strptime(
        tx['timeStamp'], "%Y-%m-%dT%H:%M:%S.%fz")
    tx_type = get_tx_type(tx['txType'])
    tx_hash = tx['txHash']

    tx = tx_types.TX(asset, tx_from, tx_to, tx_date, tx_type, tx_hash)
    return tx
